"""
This module defines the Gate class, which represents a quantum gate, as well as implementations of many common Gates.

Through the use of KroneckerGate and ProductGate, Gates can be formed for complex circuit structures.  The matrix and mat_jac functions are used for numerical optimization of parameterized gates.  The assemble function is used to generate an intermediate language of tuples that can be used by Assemblers to output descriptions of quantum circuits in other formats.
"""

import numpy as np
from . import utils, unitaries
from hashlib import md5

try:
    from qsrs import native_from_object
except ImportError:
    native_from_object = None

class Gate():
    """This class shows the framework for working with quantum gates in Qsearch."""
    def __init__(self):
        """Gates must set the following variables in __init__
        
        self.num_inputs : The number of parameters needed to generate a unitary.  This can be 0.
        self.qudits : The number of qudits acted on by a unitary of the size generated by the gate. For example, this would be 1 for U3, 2 for CNOT.
        """
        raise NotImplementedError("Subclasses of Gate should declare their own initializers.")
    
    def matrix(self, v):
        """Generates a matrix using the given vector of input parameters.  For a constant gate, v will be empty.

        Args:
            v : A numpy array of real floating point numbers, ranging from 0 to 2*PI.  Its size is equal to self.num_inputs
        
        Returns:
            np.ndarray : A unitary matrix with dtype="complex128", equal in size to d**self.qudits, where d is the intended
            qudit size (d is 2 for qubits, 3 for qutrits, etc.)
        """
        raise NotImplementedError("Subclasses of Gate are required to implement the matrix(v) method.")

    def mat_jac(self, v):
        """Generates a matrix and the jacobian(s) using the given vector of input parameters.

        It is not required to implement mat_jac for constant gates, nor is it required when using gradient-free Solvers.

        The jacobian matrices will be complex valued, and should be the elementwise partial derivative with respect to each of the parameters.
        There should be self.num_inputs matrices in the array, with the ith entry being the partial derivative with respect to v[i].
        See U3Gate for an example implementation.

        Args:
            v : A numpy array of real floating point numbers, ranging from 0 to 2*PI.  Its size is equal to self.num_inputs

        Returns:
            tuple : A tuple of the same unitary that would be returned by matrix(v), and an array of Jacobian matrices.
        """
        if self.num_inputs == 0:
            return (self.matrix(v), []) # A constant gate (one with no parameters) has no jacobian
        raise NotImplementedError("Subclasses of Gate are required to implement the mat_jac(v) method in order to be used with gradient optimizers.")

    def assemble(self, v, i=0):
        """Generates an array of tuples as an intermediate format before being processed by an Assembler for conversion to other circuit formats.

        Args:
            v : The same numpy array of real floating point numbers that might be passed to matrix(v).
            i : The index of the lowest-indexed qubit that the unitary generated by the gate acts on.


        Returns:
            list : A list of tuples following the format described above.

        The format of the tuples returned looks like:

        `("gate", gatename, (*gateparameters), (*gateindices))`

        Where `gatename` corresponds to a gate that an Assembler will recognize, `gateparameters` corresponds to the parameters for the specified gate
        (usually but not always calculated from v), and `gateindices` corresponds to the qubit indices that the gate acts on (usually but not always calculated from i).

        You can also have tuples of the form ("block", *tuples)
        Where tuples is an array of tuples in this same format.

        For some helpful examples, look at U3Gate, XZXZGate, CNOTGate, and NonadjacentCNOTGate.
        """
        raise NotImplementedError("Subclasses of Gate are required to implement the assemble(v, i) method.")

    def __eq__(self, other):
        return repr(self) == repr(other)

    def __hash__(self):
        return int(md5(repr(self).encode()).hexdigest(), 16)

    def copy(self):
        return self

    def _parts(self):
        return [self]

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    def __repr__(self):
        return "Gate()"

    def validate_structure(self):
        return True

class IdentityGate(Gate):
    """Represents an identity gate of any number of qudits of any size."""
    def __init__(self, qudits=1, d=2):
        """
        Args:
            qudits : The number of qudits represented by this identity.
            d : The size of qudits represented by this identity (2 for qubits, 3 for qutrits, etc.)
        """
        self.num_inputs=0
        self._I = np.array(np.eye(d**qudits), dtype='complex128')
        self.qudits = qudits
        self._d = d

    def matrix(self, v):
        return self._I

    def assemble(self, v, i=0):
        return []

    def __repr__(self):
        if self.qudits == 1 and self._d == 2:
            return "IdentityGate()"
        else:
            return "IdentityGate(qudits={}, d={})".format(self.qudits, self._d)


class XGate(Gate):
    """Represents a parameterized X rotation on one qubit."""
    def __init__(self):
        self.num_inputs = 1
        self.qudits = 1

    def matrix(self, v):
        return unitaries.rot_x(v[0])

    def mat_jac(self, v):
        U = unitaries.rot_x(v[0])
        J1 = unitaries.rot_x_jac(v[0])
        return U, [J1]

    def assemble(self, v, i=0):
        return [("gate", "X", (v[0],), (i,))]

    def __repr__(self):
        return "XGate()"

class YGate(Gate):
    """Represents a parameterized Y rotation on one qubit."""
    def __init__(self):
        self.num_inputs = 1
        self.qudits = 1

    def matrix(self, v):
        return unitaries.rot_y(v[0])

    def mat_jac(self, v):
        U = unitaries.rot_y(v[0])
        J1 = unitaries.rot_y_jac(v[0])
        return U, [J1]

    def assemble(self, v, i=0):
        return [("gate", "Y", (v[0],), (i,))]

    def __repr__(self):
        return "YGate()"

class ZGate(Gate):
    """Represents a parameterized Z rotation on one qubit."""
    def __init__(self):
        self.num_inputs = 1
        self.qudits = 1

    def matrix(self, v):
        return unitaries.rot_z(v[0])

    def mat_jac(self, v):
        U = unitaries.rot_z(v[0])
        J1 = unitaries.rot_z_jac(v[0])
        return U, [J1]

    def assemble(self, v, i=0):
        return [("gate", "Z", (v[0],), (i,))]

    def __repr__(self):
        return "ZGate()"

class ZXZXZGate(Gate):
    """Represents an arbitrary parameterized single-qubit gate, decomposed into 3 parameterized Z gates separated by X(PI/2) gates."""
    def __init__(self):
        self.num_inputs = 3
        self.qudits = 1

        self._x90 = unitaries.rot_x(np.pi/2)
        self._rot_z = unitaries.rot_z(0)
        self._out = np.array(np.eye(2), dtype='complex128')
        self._buffer = np.array(np.eye(2), dtype = 'complex128')
       
    def matrix(self, v):
        utils.re_rot_z(v[0], self._rot_z)
        self._out = np.dot(self._x90, self._rot_z, out=self._out)
        utils.re_rot_z(v[1], self._rot_z)
        self._buffer = np.dot(self._rot_z, self._out, out=self._buffer)
        self._out = np.dot(self._x90, self._buffer, out=self._out)
        utils.re_rot_z(v[2], self._rot_z)
        return np.dot(self._rot_z, self._out)

    def mat_jac(self, v):
        utils.re_rot_z_jac(v[0], self._rot_z)
        self._out = np.dot(self._x90, self._rot_z, out=self._out)
        utils.re_rot_z(v[1], self._rot_z)
        self._buffer = np.dot(self._rot_z, self._out, out=self._buffer)
        self._out = np.dot(self._x90, self._buffer, out=self._out)
        utils.re_rot_z(v[2], self._rot_z)
        J1 = np.dot(self._rot_z, self._out)

        utils.re_rot_z(v[0], self._rot_z)
        self._out = np.dot(self._x90, self._rot_z, out=self._out)
        utils.re_rot_z_jac(v[1], self._rot_z)
        self._buffer = np.dot(self._rot_z, self._out, out=self._buffer)
        self._out = np.dot(self._x90, self._buffer, out=self._out)
        utils.re_rot_z(v[2], self._rot_z)
        J2 = np.dot(self._rot_z, self._out)

        utils.re_rot_z(v[0], self._rot_z)
        self._out = np.dot(self._x90, self._rot_z, out=self._out)
        utils.re_rot_z(v[1], self._rot_z)
        self._buffer = np.dot(self._rot_z, self._out, out=self._buffer)
        self._out = np.dot(self._x90, self._buffer, out=self._out)
        utils.re_rot_z_jac(v[2], self._rot_z)
        J3 = np.dot(self._rot_z, self._out)
        
        utils.re_rot_z(v[2], self._rot_z)
        U = np.dot(self._rot_z, self._out)
        return (U, [J1, J2, J3])

    def assemble(self, v, i=0):
        out = []
        v = np.array(v)%(2*np.pi) # confine the range of what we print to come up with nicer numbers at no loss of generality
        out.append(("gate", "Z", (v[0],), (i,)))
        out.append(("gate", "X", (np.pi/2,), (i,)))
        out.append(("gate", "Z", (v[1],), (i,)))
        out.append(("gate", "X", (np.pi/2,), (i,)))
        out.append(("gate", "Z", (v[2],), (i,)))
        return [("block", out)]
 
    def __repr__(self):
        return "ZXZXZGate()"

class XZXZGate(Gate):
    """Represents a partially parameterized single qubit gate, equivalent to ZXZXZ but without the first Z gate.  This is useful because that first Z gate can commute through the control of a CNOT, thereby reducing the number of parameters we need to solve for."""
    def __init__(self):
        self.num_inputs = 2
        self.qudits = 1

        self._x90 = unitaries.rot_x(np.pi/2)
        self._rot_z = unitaries.rot_z(0)
        self._out = np.array(np.eye(2), dtype='complex128')
        self._buffer = np.array(np.eye(2), dtype = 'complex128')
        # need two buffers due to a bug in some implementations of numpy
        
    def matrix(self, v):
        utils.re_rot_z(v[0], self._rot_z)
        self._buffer = np.dot(self._rot_z, self._x90, out=self._buffer)
        self._out = np.dot(self._x90, self._buffer, out=self._out)
        utils.re_rot_z(v[1], self._rot_z)
        return np.dot(self._rot_z, self._out)

    def mat_jac(self, v):
        utils.re_rot_z_jac(v[0], self._rot_z)
        self._buffer = np.dot(self._rot_z, self._x90, out=self._buffer)
        self._out = np.dot(self._x90, self._buffer, out=self._out)
        utils.re_rot_z(v[1], self._rot_z)
        J1 = np.dot(self._rot_z, self._out)

        utils.re_rot_z(v[0], self._rot_z)
        self._buffer = np.dot(self._rot_z, self._x90, out=self._buffer)
        self._out = np.dot(self._x90, self._buffer, out=self._out)
        utils.re_rot_z_jac(v[1], self._rot_z)
        J2 = np.dot(self._rot_z, self._out)

        utils.re_rot_z(v[1], self._rot_z)
        U = np.dot(self._rot_z, self._out)
        return (U, [J1, J2])

    def assemble(self, v, i=0):
        out = []
        v = np.array(v)%(2*np.pi) # confine the range of what we print to come up with nicer numbers at no loss of generality
        out.append(("gate", "X", (np.pi/2,), (i,)))
        out.append(("gate", "Z", (v[0],), (i,)))
        out.append(("gate", "X", (np.pi/2,), (i,)))
        out.append(("gate", "Z", (v[1],), (i,)))
        return [("block", out)]
 
    def __repr__(self):
        return "XZXZGate()"

class U3Gate(Gate):
    """Represents an arbitrary parameterized single qubit gate, parameterized in the same way as IBM's U3 gate."""
    def __init__(self):
        self.num_inputs = 3
        self.qudits = 1

    def matrix(self, v):
        ct = np.cos(v[0]/2)
        st = np.sin(v[0]/2)
        cp = np.cos(v[1])
        sp = np.sin(v[1])
        cl = np.cos(v[2])
        sl = np.sin(v[2])
        return np.array([[ct, -st * (cl + 1j * sl)], [st * (cp + 1j * sp), ct * (cl * cp - sl * sp + 1j * cl * sp + 1j * sl * cp)]], dtype='complex128')


    def mat_jac(self, v):
        ct = np.cos(v[0]/2)
        st = np.sin(v[0]/2)
        cp = np.cos(v[1])
        sp = np.sin(v[1])
        cl = np.cos(v[2])
        sl = np.sin(v[2])

        U = np.array([[ct, -st * (cl + 1j * sl)], [st * (cp + 1j * sp), ct * (cl * cp - sl * sp + 1j * cl * sp + 1j * sl * cp)]], dtype='complex128')
        J1 = np.array([[-0.5*st, -0.5*ct * (cl + 1j * sl)], [0.5*ct * (cp + 1j * sp), -0.5*st * (cl * cp - sl * sp + 1j * cl * sp + 1j * sl * cp)]], dtype='complex128')
        J2 = np.array([[0, 0], [st *(-sp + 1j * cp), ct *(cl * -sp - sl * cp + 1j * cl * cp + 1j * sl * -sp)]], dtype='complex128')
        J3 = np.array([[0, -st *(-sl + 1j * cl)], [0, ct *(-sl * cp - cl * sp + 1j * -sl * sp + 1j * cl * cp)]], dtype='complex128')
        return (U, [J1, J2, J3])

    def assemble(self, v, i=0):
        v = np.array(v)%(2*np.pi) # confine the range to nice numbers
        return [("gate", "U3", (v[0], v[1], v[2]), (i,))]

    def __eq__(self, other):
        return type(self) == type(other)

    def __repr__(self):
        return "U3Gate()"

class U2Gate(Gate):
    """Represents a parameterized single qubit gate, parameterized in the same way as IBM's U2 gate."""
    def __init__(self):
        self.num_inputs = 2
        self.qudits = 1

    def matrix(self, v):
        return 1/np.sqrt(2) * np.array([[1, -np.exp(1j * v[1])], [np.exp(1j * v[0]), np.exp(1j * (v[0] + v[1]))]])

    def mat_jac(self, v):
        initial = 1/np.sqrt(2)
        e1 = np.exp(1j * v[1])
        e2 = np.exp(1j * v[0])
        e3 = np.exp(1j * (v[0] + v[1]))
        U = initial * np.array([[1, -e1], [e2, e3]])
        J1 = initial * np.array([[0, 0], [1j * e2, 1j * e3]])
        J2 = initial * np.array([[0, -1j * e1], [0, 1j * e3]])
        return (U, [J1, J2])

    def assemble(self, v, i=0):
        out = []
        v = np.array(v)%(2*np.pi) # confine the range of what we print to come up with nicer numbers at no loss of generality
        out.append(("gate", "Z", (v[0],), (i,)))
        out.append(("gate", "Y", (np.pi/2,), (i,)))
        out.append(("gate", "Z", (v[1],), (i,)))
        return [("block", out)]

    def __eq__(self, other):
        return type(self) == type(other)

    def __repr__(self):
        return "U2Gate()"

class SingleQutritGate(Gate):
    """This gate represents an arbitrary parameterized single-qutrit gate."""
    def __init__(self):
        self.num_inputs = 8
        self.qudits = 1

    def matrix(self, v):
        # for reference see the original implementation, qt_arb_rot in utils.py, which is now deprecated
        # this was re-written to be computationally more efficient and more readable
        s1 = np.sin(v[0])
        c1 = np.cos(v[0])
        s2 = np.sin(v[1])
        c2 = np.cos(v[1])
        s3 = np.sin(v[2])
        c3 = np.cos(v[2])
        
        p1 = np.exp(1j * v[3])
        m1 = np.exp(-1j * v[3])
        p2 = np.exp(1j * v[4])
        m2 = np.exp(-1j * v[4])
        p3 = np.exp(1j * v[5])
        m3 = np.exp(-1j * v[5])
        p4 = np.exp(1j * v[6])
        m4 = np.exp(-1j * v[6])
        p5 = np.exp(1j * v[7])
        m5 = np.exp(-1j * v[7])

        return np.array([
            [c1*c2*p1, s1*p3, c1*s2*p4],
            [s2*s3*m4*m5 - s1*c2*c3*p1*p2*m3, c1*c3*p2, -c2*s3*m1*m5 - s1*s2*c3*p2*m3*p4],
            [-s1*c2*s3*p1*m3*p5 - s2*c3*m2*m4, c1*s3*p5, c2*c3*m1*m2 - s1*s2*s3*m3*p4*p5]
            ], dtype = 'complex128')

    def mat_jac(self, v):
        s1 = np.sin(v[0])
        c1 = np.cos(v[0])
        s2 = np.sin(v[1])
        c2 = np.cos(v[1])
        s3 = np.sin(v[2])
        c3 = np.cos(v[2])
        
        p1 = np.exp(1j * v[3])
        m1 = np.exp(-1j * v[3])
        p2 = np.exp(1j * v[4])
        m2 = np.exp(-1j * v[4])
        p3 = np.exp(1j * v[5])
        m3 = np.exp(-1j * v[5])
        p4 = np.exp(1j * v[6])
        m4 = np.exp(-1j * v[6])
        p5 = np.exp(1j * v[7])
        m5 = np.exp(-1j * v[7])

        U = np.array([
            [c1*c2*p1, s1*p3, c1*s2*p4],
            [s2*s3*m4*m5 - s1*c2*c3*p1*p2*m3, c1*c3*p2, -c2*s3*m1*m5 - s1*s2*c3*p2*m3*p4],
            [-s1*c2*s3*p1*m3*p5 - s2*c3*m2*m4, c1*s3*p5, c2*c3*m1*m2 - s1*s2*s3*m3*p4*p5]
            ], dtype = 'complex128')

        Jt1 = np.array([
            [-s1*c2*p1, c1*p3, -s1*s2*p4],
            [-c1*c2*c3*p1*p2*m3, -s1*c3*p2, -c1*s2*c3*p2*m3*p4],
            [-c1*c2*s3*p1*m3*p5, -s1*s3*p5, -c1*s2*s3*m3*p4*p5]
            ], dtype = 'complex128')

        Jt2 = np.array([
            [-c1*s2*p1, 0, c1*c2*p4],
            [c2*s3*m4*m5 + s1*s2*c3*p1*p2*m3, 0, s2*s3*m1*m5 - s1*c2*c3*p2*m3*p4],
            [s1*s2*s3*p1*m3*p5 -c2*c3*m2*m4, 0, -s2*c3*m1*m2 - s1*c2*s3*m3*p4*p5]
            ], dtype = 'complex128')

        Jt3 = np.array([
            [0, 0, 0],
            [s2*c3*m4*m5 + s1*c2*s3*p1*p2*m3, -c1*s3*p2, -c2*c3*m1*m5 + s1*s2*s3*p2*m3*p4],
            [-s1*c2*c3*p1*m3*p5 + s2*s3*m2*m4, c1*c3*p5, -c2*s3*m1*m2 - s1*s2*c3*m3*p4*p5]
            ], dtype = 'complex128')

        Je1 = np.array([
            [1j*c1*c2*p1, 0, 0],
            [-1j*s1*c2*c3*p1*p2*m3, 0, 1j*c2*s3*m1*m5],
            [-1j*s1*c2*s3*p1*m3*p5, 0, -1j*c2*c3*m1*m2]
            ], dtype = 'complex128')

        Je2 = np.array([
            [0, 0, 0],
            [-1j*s1*c2*c3*p1*p2*m3, 1j*c1*c3*p2, -1j*s1*s2*c3*p2*m3*p4],
            [1j*s2*c3*m2*m4, 0, -1j*c2*c3*m1*m2]
            ], dtype = 'complex128')

        Je3 = np.array([
            [0, 1j*s1*p3, 0],
            [1j*s1*c2*c3*p1*p2*m3, 0, 1j*s1*s2*c3*p2*m3*p4],
            [1j*s1*c2*s3*p1*m3*p5, 0, 1j*s1*s2*s3*m3*p4*p5]
            ], dtype = 'complex128')

        Je4 = np.array([
            [0, 0, 1j*c1*s2*p4],
            [-1j*s2*s3*m4*m5, 0, -1j*s1*s2*c3*p2*m3*p4],
            [1j*s2*c3*m2*m4, 0, -1j*s1*s2*s3*m3*p4*p5]
            ], dtype = 'complex128')

        Je5 = np.array([
            [0, 0, 0],
            [-1j*s2*s3*m4*m5, 0, 1j*c2*s3*m1*m5],
            [-1j*s1*c2*s3*p1*m3*p5, 1j*c1*s3*p5, -1j*s1*s2*s3*m3*p4*p5]
            ], dtype = 'complex128')

        return (U, [Jt1, Jt2, Jt3, Je1, Je2, Je3, Je4, Je5])

    def assemble(self, v, i=0):
        return [("gate", "QUTRIT", (*v,), (i,))]
    
    def __repr__(self):
        return "SingleQutritGate()"

class CSUMGate(Gate):
    """Represents the constant two-qutrit gate CSUM"""
    _csum =  np.array([[1,0,0, 0,0,0, 0,0,0],
                        [0,1,0, 0,0,0, 0,0,0],
                        [0,0,1, 0,0,0, 0,0,0],
                        [0,0,0, 0,0,1, 0,0,0],
                        [0,0,0, 1,0,0, 0,0,0],
                        [0,0,0, 0,1,0, 0,0,0],
                        [0,0,0, 0,0,0, 0,1,0],
                        [0,0,0, 0,0,0, 1,0,0],
                        [0,0,0, 0,0,0, 0,0,1]
                       ], dtype='complex128')
    
    def __init__(self):
        self.num_inputs = 0
        self.qudits = 2

    def matrix(self, v):
        return CSUMGate._csum

    def assemble(self, v, i=0):
        return [("gate", "CSUM", (), (i, i+1))]

    def __repr__(self):
        return "CSUMGate()"

class CPIGate(Gate):
    """Represents the constant two-qutrit gate CPI."""
    _cpi = np.array([[1,0,0, 0,0,0, 0,0,0],
                      [0,1,0, 0,0,0, 0,0,0],
                      [0,0,1, 0,0,0, 0,0,0],
                      [0,0,0, 0,1,0,0,0,0],
                      [0,0,0, 1,0,0, 0,0,0],
                      [0,0,0, 0,0,1, 0,0,0],
                      [0,0,0, 0,0,0, 1,0,0],
                      [0,0,0, 0,0,0, 0,1,0],
                      [0,0,0, 0,0,0, 0,0,1]
                     ], dtype='complex128')
    
    def __init__(self):
        self.num_inputs = 0
        self.qudits = 2

    def matrix(self, v):
        return CPIGate._cpi

    def assemble(self, v, i=0):
        return [("gate", "CPI", (), (i, i+1))]

    def __repr__(self):
        return "CPIGate()"

class CPIPhaseGate(Gate):
    """Represents the constant two-qutrit gate CPI with phase differences."""
    def __init__(self):
        self.num_inputs = 0
        self._cpi = np.array([[1,0,0, 0,0,0, 0,0,0],
                               [0,1,0, 0,0,0, 0,0,0],
                               [0,0,1, 0,0,0, 0,0,0],
                               [0,0,0, 0,-1,0,0,0,0],
                               [0,0,0, 1,0,0, 0,0,0],
                               [0,0,0, 0,0,1, 0,0,0],
                               [0,0,0, 0,0,0, 1,0,0],
                               [0,0,0, 0,0,0, 0,1,0],
                               [0,0,0, 0,0,0, 0,0,1]
                              ], dtype='complex128')
        diag_mod = np.array(np.diag([1]*4 + [np.exp(2j * np.random.random()*np.pi) for _ in range(0,5)]))
        self._cpi = np.matmul(self._cpi, diag_mod)
        self.qudits = 2

    def matrix(self, v):
        return self._cpi

    def assemble(self, v, i=0):
        return [("gate", "CPI-", (), (i, i+1))]

    def __repr__(self):
        return "CPIPhaseGate()"

class CNOTGate(Gate):
    """Represents the constant two-qubit gate CNOT."""
    _cnot = np.array([[1,0,0,0],
                       [0,1,0,0],
                       [0,0,0,1],
                       [0,0,1,0]], dtype='complex128')
    def __init__(self):
        self.num_inputs = 0
        self.qudits = 2

    def __eq__(self, other):
        return type(self) == type(other)

    def matrix(self, v):
        return CNOTGate._cnot

    def assemble(self, v, i=0):
        return [("gate", "CNOT", (), (i, i+1))]

    def __repr__(self):
        return "CNOTGate()"

class NonadjacentCNOTGate(Gate):
    """Represents the two-qubit gate CNOT, but between two qubits that are not necessarily next to each other."""
    def __init__(self, qudits, control, target):
        """
        Args:
            qudits : The total number of qubits that a unitary of the size returned by this gate would represent.  For this gate, usually
                this is the total number of qubits in the larger circuit.
            control : The index of the control qubit, relative to the 0th qubit that would be affected by the unitary returned by this gate.
            target : The index of the target qubit, relative to the 0th qubit that would be affected by the unitary returned by this gate.
        """
        self.qudits = qudits
        self.num_inputs = 0
        self.control = control
        self.target = target
        self._U = unitaries.arbitrary_cnot(qudits, control, target)

    def matrix(self, v):
        return self._U

    def assemble(self, v, i=0):
        return [("gate", "CNOT", (), (self.control, self.target))]

    def __repr__(self):
        return "NonadjacentCNOTGate({}, {}, {})".format(self.qudits, self.control, self.target)

    def validate_structure(self):
        if self.control >= self.qudits or self.target > self.qudits:
            warn("Invalid NonadjacentCNOTGate; both control and target must be smaller than dits.  Expected {} > {} and {}".format(self.qudits, self.control, self.target))
            return False
        if self.control == self.target:
            warn("Invalid NonadjacentCNOTGate: control and target must be different. Expected {} != {}".format(self.control, self.target))
            return False
        return True

class UGate(Gate):
    """Represents an arbitrary constant gate, defined by the unitary passed to the initializer."""
    def __init__(self, U, d=2, gatename="CUSTOM", gateparams=(), gateindices=None):
        """        
        Args:
            U : The unitary for the operation that this gate represents, as a numpy ndarray with datatype="complex128".
            d : The size of qudits for the operation that this gate represents.  The default is 2, for qubits.
            gatename : A name for this gate, which will get passed to the Assembler at assembly time.
            gateparams : A tuple of parameters that will get passed to the Assembler at assembly time.
            gateindices : A tuple of indices for the qubits that this gate acts on, which will get passed to the Assembler at assembly time.  This overrides the default behavior, which is to return a tuple of all the indices starting with the one passed in assemble(v, i), and ending at i+self.qudits 
        """
        self.d = d
        self.U = U
        self.qudits = int(np.log(U.shape[0])/np.log(2))
        self.num_inputs = 0
        self.gatename = gatename
        self.gateparams = gateparams
        self.gateindices = gateindices

    def matrix(self, v):
        return self.U

    def assemble(self, v, i=0):
        gatename = self.gatename
        gateparams = self.gateparams
        indices = self.gateindices if self.gateindices is not None else tuple((i+j for j in range(self.qudits)))
        return [("gate", gatename, gateparams, indices)]

    def __repr__(self):
        if self.d == 2:
            return "UGate({})".format(repr(U))
        else:
            return "UGate({}, d={})".format(repr(U), self.d)

class UpgradedConstantGate(Gate):
    """Represents a constant gate, based on the Gate passed to its initializer, but upgraded to act on qudits of a larger size."""
    def __init__(self, other, df=3):
        """
        Args:
            other : A Gate of a lower qudit size.
            df : The final, upgraded qudit size.  The default is 3, for upgrading gates from qubits to qutrits.
        """
        if other.num_inputs > 0:
            raise AttributeError("UpgradedConstantGate is designed for only constant gates")
        OU = other.matrix([])
        di = int(OU.shape[0]**(1/other.qudits))
        if df <= di:
            raise AttributeError("Gate cannot be upgraded because it is already of an equal or higher dit level")
        self.df = df
        self.qudits = other.qudits
        self.U = utils.upgrade_qudits(OU, di, df)
        self.num_inputs = 0
        self.subgate = other

    def matrix(self, v):
        return self.U

    def assemble(self, v, i=0):
        return self.subgate.assemble(v, i)

    def __repr__(self):
        return "UpgradedConstantGate({}, df={})".format(repr(self.subgate), self.df)


class CUGate(Gate):
    """Represents an arbitrary controlled gate, defined by the unitary passed to the initializer."""
    def __init__(self, U, gatename="Name", gateparams=(), flipped=False):
        """
        Args:
            U : The unitary to form the controlled-unitary gate, in the form of a numpy ndarray with dtype="complex128"
            gatename : A name for this controlled gate which will get passed to the Assembler at assembly time.
            gateparams : A tuple of parameters that will get passed to the Assembler at assembly time.
            flipped : A boolean flag, which if set to true, will flip the direction of the gate.  The default direction is for the control qubit to be the lower indexed qubit.
        """
        self.gatename = gatename
        self.gateparams = gateparams
        self.flipped = flipped
        self.num_inputs = 0
        self._U = U
        n = np.shape(U)[0]
        I = np.array(np.eye(n))
        top = np.pad(self._U if flipped else I, [(0,n),(0,n)], 'constant')
        bot = np.pad(I if flipped else self._U, [(n,0),(n,0)], 'constant')
        self._CU = np.array(top + bot)
        self.qudits = 2
        self.num_inputs = 0

    def matrix(self, v):
        return self._CU

    def assemble(self, v, i=0):
        gatename = self.gatename
        gateparams = self.gateparams
        indices = (i, i+1) if not flipped else (i+1, i)
        return [("gate", gatename, gateparams, indices)]

    def __repr__(self):
        return "CUGate(" + str(repr(self._U)) + ("" if self.name is None else ", name={}".format(repr(self.name))) + ("flipped=True" if self.flipped else "") + ")"

class CNOTRootGate(Gate):
    """Represents the sqrt(CNOT) gate.  Two sqrt(CNOT) gates in a row will form a CNOT gate."""
    _cnr = np.array([[1,0,0,0],
                       [0,1,0,0],
                       [0,0,0.5+0.5j,0.5-0.5j],
                       [0,0,0.5-0.5j,0.5+0.5j]])
    def __init__(self):
        self.num_inputs = 0
        self.qudits = 2

    def matrix(self, v):
        return CNOTRootGate._cnr

    def assemble(self, v, i=0):
        return [("gate", "sqrt(CNOT)", (), (i, i+1))]

    def __repr__(self):
        return "CNOTRootGate()"

class KroneckerGate(Gate):
    """Represents the Kronecker product of a list of gates.  This is equivalent to performing those gate in parallel in a quantum circuit."""
    def __init__(self, *subgates):
        """
        Args:
            *subgates : An sequence of Gates.  KroneckerGate will return the kronecker product of the unitaries returned by those Gates.
        """
        self.num_inputs = sum([gate.num_inputs for gate in subgates])
        self._subgates = subgates
        self.qudits = sum([gate.qudits for gate in subgates])

    def matrix(self, v):
        if len(self._subgates) < 2:
            return self._subgates[0].matrix(v)
        matrices = []
        index = 0
        for gate in self._subgates:
            U = gate.matrix(v[index:index+gate.num_inputs])
            matrices.append(U)
            index += gate.num_inputs
        U = matrices[0]
        for matrix in matrices[1:]:
            U = np.kron(U, matrix)
        return U

    def mat_jac(self, v):
        if len(self._subgates) < 2:
            return self._subgates[0].mat_jac(v)
        matjacs = []
        index = 0
        for gate in self._subgates:
            MJ = gate.mat_jac(v[index:index+gate.num_inputs])
            matjacs.append(MJ)
            index += gate.num_inputs

        U = None
        jacs = []
        for M, Js in matjacs:
            jacs = [np.kron(J, M) for J in jacs]
            for J in Js:
                jacs.append(J if U is None else np.kron(U,J))
            U = M if U is None else np.kron(U, M)

        return (U, jacs)

    def assemble(self, v, i=0):
        out = []
        index = 0
        for gate in self._subgates:
            out += gate.assemble(v[index:index+gate.num_inputs], i)
            index += gate.num_inputs
            i += gate.qudits
        return [("block", out)]

    def appending(self, gate):
        """Returns a new KroneckerGate with the new gate added to the list.
        
        
        Args:
            gate : A Gate to be added to the end of the list of gates in the new KroneckerGate.
        """
        return KroneckerGate(*self._subgates, gate)

    def _parts(self):
        return self._subgates

    def __deepcopy__(self, memo):
        return KroneckerGate(self._subgates.__deepcopy__(memo))

    def __repr__(self):
        return "KroneckerGate({})".format(repr(self._subgates)[1:-1])

    def validate_structure(self):
        valid = True
        num_inputs = 0
        dits = 0
        for subgate in self._subgates:
            if not subgate.validate_structure():
                valid = False
            num_inputs += subgate.num_inputs
            dits += subgate.qudits

        if num_inputs != self.num_inputs:
            warn("KroneckerGate had a num_inputs mismatch: expected {} but got {}".format(self.num_inputs, num_inputs))
            valid = False
        if dits != self.qudits:
            warn("KroneckerGate had a dits mismatch: expected {} but got {}".format(self.qudits, dits))
            valid = False
        return valid

class ProductGate(Gate):
    """Represents a matrix product of Gates.  This is equivalent to performing those gates sequentially in a quantum circuit."""
    def __init__(self, *subgates):
        """
        Args:
            subgates : A list of Gates to be multiplied together.  ProductGate returns the matrix product of the unitaries returned by those Gates.
        """
        self.num_inputs = sum([gate.num_inputs for gate in subgates])
        self._subgates = []
        for subgate in subgates:
            if type(subgate) is ProductGate:
                self._subgates.extend(subgate._subgates)
            else:
                self._subgates.append(subgate)
        self.qudits = 0 if len(subgates) == 0 else subgates[0].qudits

    def matrix(self, v):
        if len(self._subgates) < 2:
            return self._subgates[0].matrix(v)
        matrices = []
        index = 0
        for gate in self._subgates:
            U = gate.matrix(v[index:index+gate.num_inputs])
            matrices.append(U)
            index += gate.num_inputs
        U = matrices[0]
        buffer1 = U.copy()
        buffer2 = U.copy()
        for matrix in matrices[1:]:
            U = np.matmul(matrix, U, out=buffer1)
            buffertmp = buffer2
            buffer2 = buffer1
            buffer1 = buffer2
        return U

    def mat_jac(self, v):
        if len(self._subgates) < 2:
            return self._subgates[0].mat_jac(v)
        submats = []
        subjacs = []
        index = 0
        for gate in self._subgates:
            U, Js = gate.mat_jac(v[index:index+gate.num_inputs])
            submats.append(U)
            subjacs.append(Js)
            index += gate.num_inputs
        
        B = np.eye(submats[0].shape[0], dtype='complex128')
        A = submats[0]
        jacs = []
        ba1 = A.copy()
        ba2 = A
        bb1 = B.copy()
        bb2 = B
        bj = B.copy()
        for matrix in submats[1:]:
            A = np.matmul(matrix, A, out=ba1)
            buffertmp = ba2
            ba2 = ba1
            ba1 = ba2

        for i, Js in enumerate(subjacs):
            A = np.matmul(A, submats[i].T.conjugate(), out=ba1) # remove the current matrix from the "after" array
            for J in Js:
                tmp = np.matmul(J, B, out=bj)
                jacs.append(np.matmul(A, tmp, out=J))
            B = np.matmul(submats[i], B, out=bb1) # add the current matrix to the "before" array before progressing
            buffertmp = ba1
            ba1 = ba2
            ba2 = buffertmp
            buffertmp = bb1
            bb1 = bb2
            bb2 = buffertmp
            
        return (B, jacs)

    def assemble(self, v, i=0):
        out = []
        index = 0
        for gate in self._subgates:
            out += gate.assemble(v[index:index+gate.num_inputs], i)
            index += gate.num_inputs
        return out

    def appending(self, *gates):
        """Returns a new ProductGate with the new gates appended to the end.

        Args:
            gates : A list of Gates to be appended.
        """
        return ProductGate(*self._subgates, *gates)

    def inserting(self, *gates, depth=-1):
        """Returns a new ProductGate with new `gates` inserted at some index `depth`.
        
        Args:
            gates : A list of Gates to be inserted.
            depth : An index in the subgates of the ProductGate after which the new gates will be inserted.  The default value of -1 will insert these gates at the begining of the ProductGate.
        """
        return ProductGate(*self._subgates[:depth], *gates, *self._subgates[depth:])

    def __deepcopy__(self, memo):
        return ProductGate(self._subgates.__deepcopy__(memo))

    def __repr__(self):
        return "ProductGate({})".format(repr(self._subgates)[1:-1])

    def validate_structure(self):
        valid = True
        num_inputs = 0
        for subgate in self._subgates:
            if subgate.qudits != self.qudits:
                warn("ProductGate had a size mismatch: expected {} but got {}".format(self.qudits, subgate.qudits))
                valid = False
            if not subgate.validate_structure():
                valid = False
            num_inputs += subgate.num_inputs

        if num_inputs != self.num_inputs:
            warn("ProductGate had a num_inputs mismatch: expected {} but got {}".format(self.num_inputs, num_inputs))
            valid = False
        valid = True



"""
This module defines the Compiler class, which is a framework for classes that take a unitary and return a circuit implementing that unitary.

The default implementation SearchCompiler is also defined here.  SearchCompiler compiles the desired unitary using an A* search strategy, as described in the paper Towards Optimal Topology Aware Quantum Circuit Synthesis.
"""
from functools import partial
from timeit import default_timer as timer
import heapq

from .gates import *

from . import solvers as scsolver
from .options import Options
from .defaults import standard_defaults, standard_smart_defaults
from . import parallelizers, backends
from . import utils, heuristics, gates, logging, gatesets

class Compiler():
    """This class defines the pattern for compilers that convert a unitary matrix to a circuit that implements that matrix."""
    def __init__(self, options=Options()):
        self.options = options

    def compile(self, options):
        raise NotImplementedError("Subclasses of Compiler are expected to implement the compile method.")
        return (U, None)

class SearchCompiler(Compiler):
    """This Compiler uses an A* search strategy to synthesize a unitary, as described in the paper Towards Optimal Topology Aware Quantum Circuit Synthesis.

    Options:
        target (required) : The unitary matrix to be synthesized, in the form of a numpy ndarray with dtype="complex128".
        gateset : The Gateset used for synthesis.
        weight_limit : A limit on the maximum weight for circuits to be expanded for further searching.  See gatesets.py for more information.  The default is None for unlimited.
        heuristic : A heuristic used to order the search tree.  See heuristics.py for more information.
        solver : A Solver used for optimizing the parameters in parameterized circuits generated by the search tree.
        parallelizer : A Parallelizer used for solving multiple parameterized circuits in parallel.
        beams : The number of nodes to pop from the search tree at a time.  The default value of -1 will create enough branches to maximize utilization of your CPU.
        error_func : The function that the Solver will attempt to minimize.
        eval_func : The function used by the heuristic in order to guide the search tree.  By default this is equal to error_func.
        error_jac : A function that returns a tuple of the value that error_func would generate and the jacobian of error_func
        error_residuals : A function that returns an array of real-valued residuals to be used by a least-squares-based Solver.
        error_residuals_jac : A function that returns the jacobian of error_residuals (note that it does NOT return a tuple of the residuals and the jacobian).
        timeout : An uper limit on the amount of time the compiler will spend trying to synthesize a circuit.  The default is float('inf'), for unlimited.
        checkpoint : The compiler will use this Checkpoint to save intermediate state, and will resume from this Checkpoint if there was an existing state.
        logger : A qsearch.logging.Logger that will be used for logging the synthesis process.

    """
    def __init__(self, options=Options()):
        """
        Args:
            options: See class level documentation for the options SearchCompiler uses
        """
        self.options = options
        self.options.set_defaults(**standard_defaults)
        self.options.set_defaults(verbosity=1, logfile=None, stdout_enabled=True)
        self.options.set_smart_defaults(**standard_smart_defaults)

    def compile(self, options=Options()):
        """
        Args:
            options: See class level documentation for the options SearchCompiler uses
        """
        options = self.options.updated(options)
        options.make_required("target")

        if "unitary_preprocessor" in options:
            U = options.unitary_preprocessor(options.target)
        weight_limit = options.weight_limit
        checkpoint = options.checkpoint
        logger = options.logger

        starttime = timer() #NOTE because all of this setup gets included in the total time, stopping and restarting the project may lead to time durations that are not representative of the runtime under normal conditions.
        h = options.heuristic
        qudits = int(np.round(np.log(np.shape(U)[0])/np.log(options.gateset.d)))

        if options.gateset.d**qudits != np.shape(U)[0]:
            raise ValueError("The target matrix of size {} is not compatible with qudits of size {}.".format(np.shape(U)[0], self.options.gateset.d))

        I = gates.IdentityGate(d=options.gateset.d)

        initial_layer = options.gateset.initial_layer(qudits)
        branching_factor = options.gateset.branching_factor(qudits)
        if branching_factor <= 0:
            logger.logprint("This gateset has no branching factor so only an initial optimization will be run.")
            root = initial_layer
            result = options.solver.solve_for_unitary(options.backend.prepare_circuit(root, options), options)
            return (root, result[1])

        parallel = options.parallelizer(options)
        # TODO move these print statements somewhere like parallelizers possibly
        logger.logprint("There are {} processors available to Pool.".format(options.num_tasks))
        logger.logprint("The branching factor is {}.".format(branching_factor))
        beams = int(options.beams)
        if beams < 1 and branching_factor > 0:
            beams = int(options.num_tasks // branching_factor)
        if beams < 1:
            beams = 1
        if beams > 1:
            logger.logprint("The beam factor is {}.".format(beams))

        recovered_state = checkpoint.recover()
        queue = []
        best_weight = 0
        best_value = 0
        best_pair  = 0
        tiebreaker = 0
        rectime = 0
        if recovered_state == None:
            if isinstance(initial_layer, ProductGate):
                root = initial_layer
            else:
                root = ProductGate(initial_layer)
            root = ProductGate(initial_layer)
            result = options.solver.solve_for_unitary(options.backend.prepare_circuit(root, options), options)
            best_value = options.eval_func(U, result[0])
            best_pair = (root, result[1])
            logger.logprint("New best! {} at weight 0".format(best_value))
            if weight_limit == 0:
                return best_pair

            queue = [(h(*best_pair, 0, options), 0, best_value, -1, result[1], root)]
            #         heuristic      weight  distance tiebreaker parameters structure
            #             0            1      2         3         4        5
            checkpoint.save((options, queue, best_weight, best_value, best_pair, tiebreaker, timer()-starttime))
        else:
            options, queue, best_weight, best_value, best_pair, tiebreaker, rectime = recovered_state
            if options.load_error:
                logger.logprint("Failed to recover state from checkpoint.  Resolve the issue or delete the checkpoint to finish the compilation.", 0)
                raise options.load_error
            logger.logprint("Recovered state with best result {} at weight {}".format(best_value, best_weight))

        options.generate_cache() # Cache the results of smart_default settings, such as the default solver, before entering the main loop where the options will get pickled and the smart_default functions called many times because later caching won't persist cause of pickeling and multiple processes.

        while len(queue) > 0:
            if timer() - starttime > options.timeout:
                break
            if best_value < options.threshold:
                queue = []
                break
            popped = []
            for _ in range(0, beams):
                if len(queue) == 0:
                    break
                tup = heapq.heappop(queue)
                popped.append(tup)
                logger.logprint("Popped a node with score: {} at weight: {}".format((tup[2]), tup[1]), verbosity=2)

            then = timer()
            new_steps = [(successor[0], current_tup[1], successor[1]) for current_tup in popped for successor in options.gateset.successors(current_tup[5])]
            for step, result, current_weight, weight in parallel.solve_circuits_parallel(new_steps):
                current_value = options.eval_func(U, result[0])
                new_weight = current_weight + weight
                if (current_value < best_value and (best_value >= options.threshold or new_weight <= best_weight)) or (current_value < options.threshold and new_weight < best_weight):
                    best_value = current_value
                    best_pair = (step, result[1])
                    best_weight = new_weight
                    logger.logprint("New best! score: {} at weight: {}".format(best_value, new_weight))
                if weight_limit is None or new_weight < weight_limit:
                    heapq.heappush(queue, (h(step, result[1], new_weight, options), new_weight, current_value, tiebreaker, result[1], step))
                    tiebreaker+=1
            logger.logprint("Layer completed after {} seconds".format(timer() - then), verbosity=2)
            checkpoint.save((options, queue, best_weight, best_value, best_pair, tiebreaker, rectime+(timer()-starttime)))


        logger.logprint("Finished compilation at weight {} with score {} after {} seconds.".format(best_weight, best_value, rectime+(timer()-starttime)))
        parallel.done()
        return {'structure': best_pair[0], 'parameters': best_pair[1]}




"""
This module describes Backend, a class which is called before the Solver is run in order to replace a Python Qsearch circuit with a Qsearch circuit based on another implementation, such as Rust or GPU.

There are three provided Backend implementations:

Attributes:
    PythonBackend : This simply returns the Python circuit, such that Python and Numpy are used for computation.
    NativeBackend : This returns the converted circuit from native_from_object, which uses the Rust implementation of Qsearch circuits provided in the qsrs module.
    SmartDefaultBackend : This backend tries to use the native Rust backend, but if it fails to convert the circuit (such as if there are unsupported gates), it will fallback to Python rather than throwing an error.
"""

try:
    from qsrs import native_from_object
    RUST_ENABLED = True
except ImportError:
    RUST_ENABLED = False
    def native_from_object(o):
        raise Exception("Native code not installed.")
from .options import Options

class Backend():
    """This class prepares a circuit for solving, replacing a Python circuit with another implementation."""
    
    def __init__(self, options=Options()):
        self.options = options

    def prepare_circuit(self, circ, options=None):
        """This function accepts a Python Qsearch circuit and returns a Qsearch circuit with a different implementation.
        Args:
            circ : The Python Qsearch circuit to be converted.
        """
        raise NotImplementedError("Subclasses of Backend must implpement prepare_circuit")

class SmartDefaultBackend(Backend):
    """This Backend tries to use the native Rust code, but will gracefully fallback to Python if there is an issue."""

    def prepare_circuit(self, circuit, options=None):
        try:
            return native_from_object(circuit)
        except:
            return circuit

class PythonBackend(Backend):
    """This Backend will simply return the Python Qsearch circuit passed in, therefore using Python and Numpy for matrix computation."""

    def prepare_circuit(self, circuit, options=None):
        return circuit

class NativeBackend(Backend):
    """This Backend will use the native Rust implementation of Qsearch circuits for faster matrix computation."""
    def prepare_circuit(self, circuit, options=None):
        return native_from_object(circuit)



"""
This module contains methods used our implementation of the Asynchronously
Parallel Optimization Solver for finding Multiple Minima (APOSMM) method.
`https://doi.org/10.1007/s12532-017-0131-4 <https://doi.org/10.1007/s12532-017-0131-4>`_

This implementation of APOSMM was developed by Kaushik Kulkarni and Jeffrey
Larson in the summer of 2019.
"""
__all__ = ['aposmm', 'initialize_APOSMM', 'decide_where_to_start_localopt', 'update_history_dist']

import numpy as np
from scipy.spatial.distance import cdist
from math import log, gamma, pi, sqrt

STOP_TAG = 3
PERSIS_STOP = 4 
FINISHED_PERSISTENT_GEN_TAG = 12


def aposmm(H, persis_info, gen_specs, libE_info):
    """
    APOSMM coordinates multiple local optimization runs, starting from points
    which do not have a better point nearby (within a distance ``r_k``). This
    generation function uses a ``local_H`` (serving a similar purpose as ``H``
    in libEnsemble) containing the fields:

    - ``'x' [n floats]``: Parameters being optimized over
    - ``'x_on_cube' [n floats]``: Parameters scaled to the unit cube
    - ``'f' [float]``: Objective function being minimized
    - ``'local_pt' [bool]``: True if point from a local optimization run
    - ``'dist_to_unit_bounds' [float]``: Distance to domain boundary
    - ``'dist_to_better_l' [float]``: Dist to closest better local opt point
    - ``'dist_to_better_s' [float]``: Dist to closest better sample point
    - ``'ind_of_better_l' [int]``: Index of point ``'dist_to_better_l``' away
    - ``'ind_of_better_s' [int]``: Index of point ``'dist_to_better_s``' away
    - ``'started_run' [bool]``: True if point has started a local opt run
    - ``'num_active_runs' [int]``: Number of active local runs point is in
    - ``'local_min' [float]``: True if point has been ruled a local minima
    - ``'sim_id' [int]``: Row number of entry in history

    and optionally

    - ``'fvec' [m floats]``: All objective components (if performing a least-squares calculation)
    - ``'grad' [n floats]``: The gradient (if available) of the objective with respect to `x`.

    Note:
    - If any of the above fields are desired after a libEnsemble run, name
      them in ``gen_specs['out']``.
    - If intitializing APOSMM with past function values, make sure to include
      ``'x'``, ``'x_on_cube'``, ``'f'``, ``'local_pt'``, etc. in
      ``gen_specs['in']`` (and, of course, include them in the H0 array given
      to libensemble).

    Necessary quantities in ``gen_specs['user']`` are:

    - ``'lb' [n floats]``: Lower bound on search domain
    - ``'ub' [n floats]``: Upper bound on search domain
    - ``'localopt_method' [str]``: Name of an NLopt, PETSc/TAO, or SciPy method
      (see 'advance_local_run' below for supported methods)
    - ``'initial_sample_size' [int]``: Number of uniformly sampled points
      must be returned (non-nan value) before a local opt run is started. Can be
      zero if no additional sampling is desired, but if zero there must be past
      sim_f values given to libEnsemble in H0.

    Optional ``gen_specs['user']`` entries are:

    - ``'sample_points' [numpy array]``: Points to be sampled (original domain).
      If more sample points are needed by APOSMM during the course of the
      optimization, points will be drawn uniformly over the domain
    - ``'components' [int]``: Number of objective components
    - ``'dist_to_bound_multiple' [float in (0,1]]``: What fraction of the
      distance to the nearest boundary should the initial step size be in
      localopt runs
    - ``'lhs_divisions' [int]``: Number of Latin hypercube sampling partitions
      (0 or 1 results in uniform sampling)
    - ``'mu' [float]``: Distance from the boundary that all localopt starting
      points must satisfy
    - ``'nu' [float]``: Distance from identified minima that all starting
      points must satisfy
    - ``'rk_const' [float]``: Multiplier in front of the r_k value
    - ``'max_active_runs' [int]``: Bound on number of runs APOSMM is advancing

    If the rules in ``decide_where_to_start_localopt`` produces more than
    ``'max_active_runs'`` in some iteration, then existing runs are prioritized.

    And ``gen_specs['user']`` must also contain fields for the given
    localopt_method's convergence tolerances (e.g., gatol/grtol for PETSC/TAO
    or ftol_rel for NLopt)

    .. seealso::

        `test_persistent_aposmm_scipy <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_persistent_aposmm_scipy.py>`_
        for basic APOSMM usage.

    .. seealso::

        `test_persistent_aposmm_with_grad <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_persistent_aposmm_with_grad.py>`_
        for an example where past function values are given to libEnsemble/APOSMM.

    """
    """
    Description of intermediate variables in aposmm:

    n:                domain dimension
    n_s:              the number of complete evaluations of sampled points
    updated_inds:     indices of H that have been updated (and so all their
                      information must be sent back to libE manager to update)
    O:                new points to be sent back to the history


    When re-running a local opt method to get the next point:
    advance_local_run.x_new:      stores the first new point requested by
                                  a local optimization method
    advance_local_run.pt_in_run:  counts function evaluations to know
                                  when a new point is given

    starting_inds:    indices where a runs should be started.
    active_runs:      indices of active local optimization runs
    sorted_run_inds:  indices of the considered run (in the order they were
                      requested by the localopt method)
    x_opt:            the reported minimum from a localopt run (disregarded
                      unless opt_flag is 1)
    opt_flag:         1 if the run ended with an optimal point (x_opt) or
                      0 if it ended because e.g., maxiters/maxevals were reached
    num_samples_needed:   Number of additional uniformly drawn samples needed


    Description of persistent variables used to maintain the state of APOSMM

    persis_info['total_runs']: Running count of started/completed localopt runs
    persis_info['run_order']: Sequence of indices of points in unfinished runs
    persis_info['old_runs']: Sequence of indices of points in finished runs

    """
    user_specs = gen_specs['user']

    n, n_s, rk_const, ld, mu, nu, comm, local_H = initialize_APOSMM(H, user_specs, libE_info)
    local_opters, sim_id_to_child_inds, run_order, run_pts, total_runs, fields_to_pass = initialize_children(user_specs)

    if user_specs['initial_sample_size'] != 0:
        # Send our initial sample. We don't need to check that n_s is large enough:
        # the alloc_func only returns when the initial sample has function values.
        persis_info = add_k_sample_points_to_local_H(user_specs['initial_sample_size'], user_specs,
                                                     persis_info, n, comm, local_H,
                                                     sim_id_to_child_inds)
        if not user_specs.get('standalone'):
            send_mgr_worker_msg(comm, local_H[-user_specs['initial_sample_size']:][[i[0] for i in gen_specs['out']]])
        something_sent = True
    else:
        something_sent = False

    tag = None
    first_pass = True
    while 1:
        new_opt_inds_to_send_mgr = []
        new_inds_to_send_mgr = []

        if something_sent:
            if user_specs.get('standalone'):
                tag, Work, calc_in = simulate_recv_from_manager(local_H, gen_specs)
            else:
                tag, Work, calc_in = get_mgr_worker_msg(comm)

            if tag in [STOP_TAG, PERSIS_STOP]:
                clean_up_and_stop(local_H, local_opters, run_order)
                persis_info['run_order'] = run_order
                break

            n_s, n_r = update_local_H_after_receiving(local_H, n, n_s, user_specs, Work, calc_in, fields_to_pass)

            for row in calc_in:
                if sim_id_to_child_inds.get(row['sim_id']):
                    # Point came from a child local opt run
                    for child_idx in sim_id_to_child_inds[row['sim_id']]:
                        x_new = local_opters[child_idx].iterate(row[fields_to_pass])
                        if isinstance(x_new, ConvergedMsg):
                            x_opt = x_new.x
                            opt_flag = x_new.opt_flag
                            opt_ind = update_history_optimal(x_opt, opt_flag, local_H, run_order[child_idx])
                            new_opt_inds_to_send_mgr.append(opt_ind)
                            local_opters.pop(child_idx)
                        else:
                            add_to_local_H(local_H, x_new, user_specs, local_flag=1, on_cube=True)
                            new_inds_to_send_mgr.append(len(local_H)-1)

                            run_order[child_idx].append(local_H[-1]['sim_id'])
                            run_pts[child_idx].append(x_new)
                            if local_H[-1]['sim_id'] in sim_id_to_child_inds:
                                sim_id_to_child_inds[local_H[-1]['sim_id']] += (child_idx, )
                            else:
                                sim_id_to_child_inds[local_H[-1]['sim_id']] = (child_idx, )

        starting_inds = decide_where_to_start_localopt(local_H, n, n_s, rk_const, ld, mu, nu)

        for ind in starting_inds:
            if len([p for p in local_opters.values() if p.is_running]) < user_specs.get('max_active_runs', np.inf):
                local_H['started_run'][ind] = 1

                # Initialize a local opt run
                local_opter = LocalOptInterfacer(user_specs, local_H[ind]['x_on_cube'],
                                                 local_H[ind]['f'] if 'f' in fields_to_pass else local_H[ind]['fvec'],
                                                 local_H[ind]['grad'] if 'grad' in fields_to_pass else None)

                local_opters[total_runs] = local_opter

                x_new = local_opter.iterate(local_H[ind][fields_to_pass])  # Assuming the second point can't be ruled optimal

                add_to_local_H(local_H, x_new, user_specs, local_flag=1, on_cube=True)
                new_inds_to_send_mgr.append(len(local_H)-1)

                run_order[total_runs] = [ind, local_H[-1]['sim_id']]
                run_pts[total_runs] = [local_H['x_on_cube'], x_new]

                if local_H[-1]['sim_id'] in sim_id_to_child_inds:
                    sim_id_to_child_inds[local_H[-1]['sim_id']] += (total_runs, )
                else:
                    sim_id_to_child_inds[local_H[-1]['sim_id']] = (total_runs, )

                total_runs += 1

        if first_pass:
            num_samples_needed = persis_info['nworkers'] - 1 - len(new_inds_to_send_mgr)
            first_pass = False
        else:
            num_samples_needed = n_r-len(new_inds_to_send_mgr)

        if num_samples_needed > 0:
            persis_info = add_k_sample_points_to_local_H(num_samples_needed, user_specs, persis_info, n, comm, local_H, sim_id_to_child_inds)
            new_inds_to_send_mgr = new_inds_to_send_mgr + list(range(len(local_H)-num_samples_needed, len(local_H)))

        if not user_specs.get('standalone'):
            send_mgr_worker_msg(comm, local_H[new_inds_to_send_mgr + new_opt_inds_to_send_mgr][[i[0] for i in gen_specs['out']]])
        something_sent = True

    return local_H, persis_info, FINISHED_PERSISTENT_GEN_TAG


def update_local_H_after_receiving(local_H, n, n_s, user_specs, Work, calc_in, fields_to_pass):

    for name in ['f', 'x_on_cube', 'grad', 'fvec', 'jac']:
        if name in fields_to_pass:
            assert name in calc_in.dtype.names, name + " must be returned to persistent_aposmm for localopt_method: " + user_specs['localopt_method']

    for name in calc_in.dtype.names:
        local_H[name][Work['libE_info']['H_rows']] = calc_in[name]

    local_H['returned'][Work['libE_info']['H_rows']] = True
    n_s += np.sum(~local_H[Work['libE_info']['H_rows']]['local_pt'])
    n_r = len(Work['libE_info']['H_rows'])

    # dist -> distance
    update_history_dist(local_H, n)

    return n_s, n_r


def add_to_local_H(local_H, pts, user_specs, local_flag=0, on_cube=True):
    """
    Adds points to O, the numpy structured array to be sent back to the manager
    """
    assert not local_flag or len(pts) == 1, "Can't > 1 local points"

    len_local_H = len(local_H)

    ub = user_specs['ub']
    lb = user_specs['lb']

    num_pts = len(pts)

    local_H.resize(len(local_H)+num_pts, refcheck=False)  # Adds num_pts rows of zeros to O

    if on_cube:
        local_H['x_on_cube'][-num_pts:] = pts
        local_H['x'][-num_pts:] = pts*(ub-lb)+lb
    else:
        local_H['x_on_cube'][-num_pts:] = (pts-lb)/(ub-lb)
        local_H['x'][-num_pts:] = pts

    if user_specs.get('periodic'):
        local_H['x_on_cube'][-num_pts:] = local_H['x_on_cube'][-num_pts:] % 1

    local_H['sim_id'][-num_pts:] = np.arange(len_local_H, len_local_H+num_pts)
    local_H['local_pt'][-num_pts:] = local_flag

    initialize_dists_and_inds(local_H, num_pts)

    if local_flag:
        local_H['num_active_runs'][-num_pts] += 1


def initialize_dists_and_inds(local_H, num_pts):
    local_H['dist_to_unit_bounds'][-num_pts:] = np.inf
    local_H['dist_to_better_l'][-num_pts:] = np.inf
    local_H['dist_to_better_s'][-num_pts:] = np.inf
    local_H['ind_of_better_l'][-num_pts:] = -1
    local_H['ind_of_better_s'][-num_pts:] = -1


def update_history_dist(H, n):
    """
    Updates distances/indices after new points that have been evaluated.

    .. seealso::
        `start_persistent_local_opt_gens.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/alloc_funcs/start_persistent_local_opt_gens.py>`_
    """

    new_inds = np.where(~H['known_to_aposmm'])[0]

    p = np.logical_and.reduce((H['returned'], ~np.isnan(H['f'])))

    for new_ind in new_inds:
        # Loop over new returned points and update their distances
        if p[new_ind]:
            H['known_to_aposmm'][new_ind] = True

            # Compute distance to boundary
            H['dist_to_unit_bounds'][new_ind] = min(min(np.ones(n)-H['x_on_cube'][new_ind]), min(H['x_on_cube'][new_ind]-np.zeros(n)))

            dist_to_all = cdist(H['x_on_cube'][[new_ind]], H['x_on_cube'][p], 'euclidean').flatten()
            new_better_than = H['f'][new_ind] < H['f'][p]

            # Update any other points if new_ind is closer and better
            if H['local_pt'][new_ind]:
                inds_of_p = np.logical_and(dist_to_all < H['dist_to_better_l'][p], new_better_than)
                updates = np.where(p)[0][inds_of_p]
                H['dist_to_better_l'][updates] = dist_to_all[inds_of_p]
                H['ind_of_better_l'][updates] = new_ind
            else:
                inds_of_p = np.logical_and(dist_to_all < H['dist_to_better_s'][p], new_better_than)
                updates = np.where(p)[0][inds_of_p]
                H['dist_to_better_s'][updates] = dist_to_all[inds_of_p]
                H['ind_of_better_s'][updates] = new_ind

            # Since we allow equality when deciding better_than_new_l and
            # better_than_new_s, we have to prevent new_ind from being its own
            # better point.
            better_than_new_l = np.logical_and.reduce((~new_better_than, H['local_pt'][p], H['sim_id'][p] != new_ind))
            better_than_new_s = np.logical_and.reduce((~new_better_than, ~H['local_pt'][p], H['sim_id'][p] != new_ind))

            # Who is closest to ind and better
            if np.any(better_than_new_l):
                ind = dist_to_all[better_than_new_l].argmin()
                H['ind_of_better_l'][new_ind] = H['sim_id'][p][np.nonzero(better_than_new_l)[0][ind]]
                H['dist_to_better_l'][new_ind] = dist_to_all[better_than_new_l][ind]

            if np.any(better_than_new_s):
                ind = dist_to_all[better_than_new_s].argmin()
                H['ind_of_better_s'][new_ind] = H['sim_id'][p][np.nonzero(better_than_new_s)[0][ind]]
                H['dist_to_better_s'][new_ind] = dist_to_all[better_than_new_s][ind]

            # if not ignore_L8:
            #     r_k = calc_rk(len(H['x_on_cube'][0]), n_s, rk_const, lhs_divisions)
            #     H['worse_within_rk'][new_ind][p] = np.logical_and.reduce((H['f'][new_ind] <= H['f'][p], dist_to_all <= r_k))

            #     # Add trues if new point is 'worse_within_rk'
            #     inds_to_change = np.logical_and.reduce((H['dist_to_all'][p,new_ind] <= r_k, H['f'][new_ind] >= H['f'][p], H['sim_id'][p] != new_ind))
            #     H['worse_within_rk'][inds_to_change,new_ind] = True

            #     if not H['local_pt'][new_ind]:
            #         H['worse_within_rk'][H['dist_to_all'] > r_k] = False

    if np.any(~H['local_pt']) and not np.any(np.isinf(H['dist_to_better_s'][~H['local_pt']])):
        # Our best sample point was not identified because the min was not unique.
        min_inds = H['f'][~H['local_pt']] == np.min(H['f'][~H['local_pt']])
        assert len(min_inds) >= 2, "Check this"
        # Take the first point with this value to be the best sample point
        best_samp = H['sim_id'][~H['local_pt']][min_inds][0]
        H['dist_to_better_s'][best_samp] = np.inf
        H['ind_of_better_s'][best_samp] = -1

    # if np.any(H['local_pt']) and not np.any(np.isinf(H['dist_to_better_l'][H['local_pt']])):
    #     # Our best sample point was not identified because the min was not unique.
    #     min_inds = H['f'][H['local_pt']] == np.min(H['f'][H['local_pt']])
    #     assert len(min_inds) >= 2, "Check this"
    #     # Take the first point with this value to be the best sample point
    #     best_local = H['sim_id'][H['local_pt']][min_inds][0]
    #     H['dist_to_better_l'][best_local] = np.inf
    #     H['ind_of_better_l'][best_local] = -1


def update_history_optimal(x_opt, opt_flag, H, run_inds):
    """
    Updated the history after any point has been declared a local minimum
    """

    # opt_ind = np.where(np.logical_and(np.equal(x_opt,H['x_on_cube']).all(1),~np.isinf(H['f'])))[0] # This fails on some problems. x_opt is 1e-16 away from the point that was given and opt_ind is empty
    run_inds = np.unique(run_inds)

    dists = np.linalg.norm(H['x_on_cube'][run_inds]-x_opt, axis=1)
    ind = np.argmin(dists)
    opt_ind = run_inds[ind]

    tol_x1 = 1e-15

    # Instead of failing, we accept x_opt that is slightly different from its value in H
    # assert dists[ind] <= tol_x1, "Closest point to x_opt not within {}?".format(tol_x1)

    if dists[ind] > tol_x1:
        print("Dist from reported x_opt to closest evaluated point is: " + str(dists[ind]) + "\n" +
              "Check that the local optimizer is working correctly\n", x_opt, run_inds, flush=True)

    tol_x2 = 1e-8
    failsafe = np.logical_and(H['f'][run_inds] < H['f'][opt_ind], dists < tol_x2)
    if opt_flag:
        if np.any(failsafe):
            print("This run has {} point(s) with smaller 'f' value within {} of "
                  "the point ruled to be the run minimum. \nMarking all as being "
                  "a 'local_min' to prevent APOSMM from starting another run "
                  "immediately from these points.".format(sum(failsafe), tol_x2))
            print("Sim_ids to be marked optimal: ", opt_ind, run_inds[failsafe])
            print("Check that the local optimizer is working correctly", flush=True)
            H['local_min'][run_inds[failsafe]] = 1

        H['local_min'][opt_ind] = 1

    H['num_active_runs'][run_inds] -= 1

    return opt_ind


def decide_where_to_start_localopt(H, n, n_s, rk_const, ld=0, mu=0, nu=0):
    """
    APOSMM starts a local optimization runs from a point that:

    - is not in an active local optimization run,
    - is more than ``mu`` from the boundary (in the unit-cube domain),
    - is more than ``nu`` from identified minima (in the unit-cube domain),
    - does not have a better point within a distance ``r_k`` of it.


    For further details, see the conditions (S1-S5 and L1-L8) in Table 1 of the
    `APOSMM paper <https://doi.org/10.1007/s12532-017-0131-4>`_
    This method first identifies sample points satisfying S2-S5, and then
    identifies all localopt points that satisfy L1-L7.
    We then start from any sample point also satisfying S1.
    We do not check condition L8 currently.

    We don't consider points in the history that have not returned from
    computation, or that have a ``nan`` value. As APOSMM works on the unit
    cube, note that ``mu`` and ``nu`` implicitly depend on the scaling of the
    original domain: adjusting the initial domain can make a run start (or not
    start) at a point that didn't (or did) previously.

    Parameters
    ----------
    H: numpy structured array
        History array storing rows for each point.
    n: int
        Problem dimension
    n_s: int
        Number of sample points in H
    r_k_const: float
        Radius for deciding when to start runs
    ld: integer
        Number of Latin hypercube sampling divisions (0 or 1 means uniform
        random sampling over the domain)
    mu: nonnegative float
        Distance from the boundary that all starting points must satisfy
    nu: nonnegative float
        Distance from identified minima that all starting points must satisfy

    Returns
    ----------
    start_inds: list
        Indices where a local opt run should be started, sorted by increasing
        function value.


    .. seealso::
        `start_persistent_local_opt_gens.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/alloc_funcs/start_persistent_local_opt_gens.py>`_
    """

    r_k = calc_rk(n, n_s, rk_const, ld)

    if nu > 0:
        test_2_through_5 = np.logical_and.reduce((
            H['returned'] == 1,  # have a returned function value
            H['dist_to_better_s'] >
            r_k,  # no better sample point within r_k (L2)
            ~H['started_run'],  # have not started a run (L3)
            H['dist_to_unit_bounds'] >=
            mu,  # have all components at least mu away from bounds (L4)
            np.all(
                cdist(H['x_on_cube'], H['x_on_cube'][H['local_min']]) >= nu,
                axis=1)  # distance nu away from known local mins (L5)
        ))
    else:
        test_2_through_5 = np.logical_and.reduce((
            H['returned'] == 1,  # have a returned function value
            H['dist_to_better_s'] >
            r_k,  # no better sample point within r_k (L2)
            ~H['started_run'],  # have not started a run (L3)
            H['dist_to_unit_bounds'] >=
            mu,  # have all components at least mu away from bounds (L4)
        ))  # (L5) is always true when nu = 0

    # assert gamma_quantile == 1, "This is not supported yet. What is the best way to decide this when there are NaNs present in H['f']?"
    # if gamma_quantile < 1:
    #     cut_off_value = np.sort(H['f'][~H['local_pt']])[np.floor(gamma_quantile*(sum(~H['local_pt'])-1)).astype(int)]
    # else:
    #     cut_off_value = np.inf

    # Find the indices of points that...
    sample_seeds = np.logical_and.reduce((
        ~H['local_pt'],  # are not localopt points
        # H['f'] <= cut_off_value,      # have a small enough objective value
        ~np.isinf(H['f']),  # have a non-infinity objective value
        ~np.isnan(H['f']),  # have a non-NaN objective value
        test_2_through_5,  # satisfy tests 2 through 5
    ))

    # Uncomment the following to test the effect of ignorning LocalOpt points
    # in APOSMM. This allows us to test a parallel MLSL.
    # return list(np.ix_(sample_seeds)[0])

    those_satisfying_S1 = H['dist_to_better_l'][sample_seeds] > r_k  # no better localopt point within r_k
    sample_start_inds = np.ix_(sample_seeds)[0][those_satisfying_S1]

    # Find the indices of points that...
    local_seeds = np.logical_and.reduce((
        H['local_pt'],  # are localopt points
        H['dist_to_better_l'] > r_k,  # no better local point within r_k (L1)
        ~np.isinf(H['f']),  # have a non-infinity objective value
        ~np.isnan(H['f']),  # have a non-NaN objective value
        test_2_through_5,
        H['num_active_runs'] == 0,  # are not in an active run (L6)
        ~H['local_min']  # are not a local min (L7)
    ))

    local_start_inds2 = list(np.ix_(local_seeds)[0])

    # If paused is a field in H, don't start from paused points.
    if 'paused' in H.dtype.names:
        sample_start_inds = sample_start_inds[~H[sample_start_inds]['paused']]
    start_inds = list(sample_start_inds)+local_start_inds2

    # Sort the starting inds by their function value
    inds = np.argsort(H['f'][start_inds])
    start_inds = np.array(start_inds)[inds].tolist()

    return start_inds


def calc_rk(n, n_s, rk_const, lhs_divisions=0):
    """ Calculate the critical distance r_k """

    if lhs_divisions == 0:
        r_k = rk_const*(log(n_s)/n_s)**(1/n)
    else:
        k = np.floor(n_s/lhs_divisions).astype(int)
        if k <= 1:  # to prevent r_k=0
            r_k = np.inf
        else:
            r_k = rk_const*(log(k)/k)**(1/n)

    return r_k


def initialize_APOSMM(H, user_specs, libE_info):
    """
    Computes common values every time that APOSMM is reinvoked

    .. seealso::
        `start_persistent_local_opt_gens.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/alloc_funcs/start_persistent_local_opt_gens.py>`_
    """
    n = len(user_specs['ub'])

    rk_c = user_specs.get('rk_const', ((gamma(1+(n/2.0))*5.0)**(1.0/n))/sqrt(pi))
    ld = user_specs.get('lhs_divisions', 0)
    mu = user_specs.get('mu', 1e-4)
    nu = user_specs.get('nu', 0)

    comm = libE_info['comm'] if not user_specs.get('standalone') else []

    local_H_fields = [('f', float),
                      ('grad', float, n),
                      ('x', float, n),
                      ('x_on_cube', float, n),
                      ('local_pt', bool),
                      ('known_to_aposmm', bool),
                      ('dist_to_unit_bounds', float),
                      ('dist_to_better_l', float),
                      ('dist_to_better_s', float),
                      ('ind_of_better_l', int),
                      ('ind_of_better_s', int),
                      ('started_run', bool),
                      ('num_active_runs', int),
                      ('local_min', bool),
                      ('sim_id', int),
                      ('paused', bool),
                      ('returned', bool),
                      ]

    if 'components' in user_specs:
        local_H_fields += [('fvec', float, user_specs['components'])]
        if 'lm_jac_func' in user_specs['standalone']:
            local_H_fields += [('jac', float, (user_specs['components'],n))]

    local_H = np.zeros(len(H), dtype=local_H_fields)

    if len(H):
        for field in H.dtype.names:
            local_H[field][:len(H)] = H[field]

        if user_specs['localopt_method'] in ['LD_MMA', 'blmvm']:
            assert 'grad' in H.dtype.names, "Must give 'grad' values to persistent_aposmm in gen_specs['in'] when using 'localopt_method'" + user_specs['localopt_method']
            assert not np.all(local_H['grad'] == 0), "All 'grad' values are zero for the given points."

        assert 'f' in H.dtype.names, "Must give 'f' values to persistent_aposmm in gen_specs['in']"
        assert 'sim_id' in H.dtype.names, "Must give 'sim_id' to persistent_aposmm in gen_specs['in']"
        assert 'returned' in H.dtype.names, "Must give 'returned' status to persistent_aposmm in gen_specs['in']"

        over_written_fields = ['dist_to_unit_bounds', 'dist_to_better_l', 'dist_to_better_s', 'ind_of_better_l', 'ind_of_better_s']
        if any([i in H.dtype.names for i in over_written_fields]):
            print("\n persistent_aposmm ignores any given values in these fields: " + str(over_written_fields) + "\n")

        initialize_dists_and_inds(local_H, len(H))

        # Update after receiving initial points
        update_history_dist(local_H, n)

    n_s = np.sum(~local_H['local_pt'])

    assert n_s > 0 or user_specs['initial_sample_size'] > 0, "APOSMM requires a positive initial_sample_size, or some existing points in order to determine where to start local optimization runs."

    if 'sample_points' in user_specs:
        assert isinstance(user_specs['sample_points'], np.ndarray)

    return n, n_s, rk_c, ld, mu, nu, comm, local_H


def initialize_children(user_specs):
    """ Initialize stuff for localopt children """
    local_opters = {}
    sim_id_to_child_inds = {}
    run_order = {}
    run_pts = {}  # This can differ from 'x_on_cube' if, for example, user_specs['periodic'] is True and run points are off the cube.
    total_runs = 0
    if user_specs['localopt_method'] in ['LD_MMA', 'blmvm', 'scipy_BFGS']:
        fields_to_pass = ['x_on_cube', 'f', 'grad']
    elif user_specs['localopt_method'] in ['LN_SBPLX', 'LN_BOBYQA', 'LN_COBYLA', 'LN_NEWUOA',
                                           'LN_NELDERMEAD', 'scipy_Nelder-Mead', 'scipy_COBYLA',
                                           'external_localopt', 'nm']:
        fields_to_pass = ['x_on_cube', 'f']
    elif user_specs['localopt_method'] in ['pounders', 'dfols']:
        fields_to_pass = ['x_on_cube', 'fvec']
    elif user_specs['localopt_method'] in ['scipy_least_squares_lm']:
        fields_to_pass = ['x_on_cube', 'fvec', 'jac']
    else:
        raise NotImplementedError("Unknown local optimization method " "'{}'.".format(user_specs['localopt_method']))

    return local_opters, sim_id_to_child_inds, run_order, run_pts, total_runs, fields_to_pass


def add_k_sample_points_to_local_H(k, user_specs, persis_info, n, comm, local_H, sim_id_to_child_inds):

    if 'sample_points' in user_specs:
        v = np.sum(~local_H['local_pt'])  # Number of sample points so far
        sampled_points = user_specs['sample_points'][v:v+k]
        on_cube = False  # Assume points are on original domain, not unit cube
        if len(sampled_points):
            add_to_local_H(local_H, sampled_points, user_specs, on_cube=on_cube)
        k = k-len(sampled_points)

    if k > 0:
        sampled_points = persis_info['rand_stream'].uniform(0, 1, (k, n))
        add_to_local_H(local_H, sampled_points, user_specs, on_cube=True)

    return persis_info


def clean_up_and_stop(local_H, local_opters, run_order):
    # FIXME: This has to be a clean exit.

    # print('[Parent]: The optimal points and values are:\n',
    #       local_H[np.where(local_H['local_min'])][['x', 'f']], flush=True)

    for i, p in local_opters.items():
        p.destroy(local_H['x_on_cube'][run_order[i][-1]])



"""
This module defines the Logger class, which is used to control and automate the logging of messages to stdout and to files.
"""
import datetime
import os.path as path
import numpy as np

class Logger:
    """This class is used to control what level of mesages get printed, and to where."""
    def __init__(self, stdout_enabled=False, output_file=None, verbosity=1):
        self.stdout_enabled=stdout_enabled
        self.output_file=output_file
        self.verbosity=verbosity

    def logprint(self, string, verbosity=1):
        """This function logs the specified string according to the specified options."""
        if verbosity > self.verbosity:
            return # ignore print requests for a higher verbosity than our current setting
        if self.stdout_enabled:
            print(string)
        if self.output_file is not None:
            with open(self.output_file, "a") as f:
                f.write(str(datetime.datetime.now()) + " > \t" + str(string) + "\n")





"""
This module defines the Checkpoint class, which is used for storing intermediate state while compiling, to allow an interrupted compilation to resume at a later time.

Two default implementations are provided.  It is recommended that you look at FileCheckpoint as an example if you are interested in writing your own implementation.

Attributes:
    FileCheckpoint : Saves and recovers the intermediate state from a file, specified as "statefile" in the options.
    ChildCheckpoint : Allows for hierarchial checkpointing, which is useful in cases where there are sub-compilers, such as with LEAP.
"""

import pickle
import os
from . import options


class Checkpoint():
    """This class is used for storing intermediate state while compiling, to allow an interrupted compilation to resume at a later time."""
    def __init__(self, options=options.Options()):
        self.options = options

    def save(self, state):
        """
        Save the passed state to be recovered later.
        Args:
            state (object): A Python object representing the intermediate state of the compilation.  Usually a dictionary, but it could be anything.
        """
        raise NotImplementedError

    def recover(self):
        """
        Return the state previously stored with save(state).

        Returns:
            object : A Python object equivalent to the object originally stored via save(state), or None if no state is saved.
        """
        raise NotImplementedError

    def delete(self):
        """
        Delete the state that was stored such that None will be returned next time recover() is called.
        """
        raise NotImplementedError


class FileCheckpoint(Checkpoint):
    """This Checkpoint will store the state in the file specified in the options as statefile.

        Options:
            statefile : A string with a filepath where the state will be stored, or None, in which case no state will be stored and None will always be returned by recover()
    """

    def __init__(self, options=options.Options()):
        super().__init__(options)
        options.set_defaults(statefile=None)

    def save(self, state):
        if self.options.statefile == None:
            return
        with open(self.options.statefile, "wb") as tmpfile:
            pickle.dump(state, tmpfile, pickle.HIGHEST_PROTOCOL)

    def recover(self):
        if self.options.statefile == None:
            return None
        try: 
            with open(self.options.statefile, "rb") as tmpfile:
                return pickle.load(tmpfile)
        except:
            return None

    def delete(self):
        if self.options.statefile == None:
            return
        try:
            os.remove(self.options.statefile)
        except:
            return

class ChildCheckpoint(Checkpoint):
    """This Checkpoint is used for hierarchial checkpointing for when there is a sub-compiler, such as in LEAP.

        Options:
            parent (required) : The Checkpoint class that the creator of the ChildCheckpoint was passed.

        Below is an explanation of how ChildCheckpoint works.  See leap_compiler for an example.

        My compiler class, ParentCompiler, is passed a FileCheckpoint as options.checkpoint.
        I create a ChildCheckpoint with the FileCheckpoint as the parent: child_checkpoint = ChildCheckpoint(Options(parent=options.checkpoint))
        I pass this ChildCheckpoint to the sub-compiler I create: sub_compiler = SubCompiler(Options(checkpoint=child_checkpoint))
        
        To the SubCompiler, the passed ChildCheckpoint will behave as any other Checkpoint would be expected to behavior, saving state with save(state), and recovering it with recover(), and deleting it with delete().

        As the ParentCompiler, you save your state with save_parent(parentstate), recover it with recover_parent(), and deleting it with delete_parent(), making these function calls to child_checkpoint instead of interacting directly with the FileCheckpoint that was originally passed via options.

        The states of both ParentCompiler and SubCompiler will get saved in a manner specified by the original FileCheckpoint.

        ChildCheckpoint fully conforms to Checkpoint, and makes no assumptions about its parent, so it is compatible with any class that makes use ofa Checkpoint and works with any Checkpoint as a parent.  This means you can even have multiple layers of nested ChildCheckpoint.

        However, the class creating the ChildCheckpoint must be sure to use the parent functions.

        Also, note that calling delete_parent() also deletes the state for the child.  However, this is rather uncommon because usually it is the creator of the Checkpoint that calls delete, not the class it is passed to.  For example, Project will call delete() to delete the checkpoint from a Compiler.  ParentCompiler might call delete() to delete the state of SubCompiler once SubCompiler has finished (in fact, this happens in leap_compiler).
    """
    def __init__(self, options=options.Options()):
        super().__init__(options)
        self.options.make_required("parent")
        self.recover()

    def save(self, state):
        self.state = state
        self.options.parent.save((self.parentstate, self.state))

    def save_parent(self, parentstate):
        """Saves the parentstate alongside the child state."""
        self.parentstate = parentstate
        self.save(self.state)

    def recover(self):
        recdata = self.options.parent.recover()
        if recdata is None:
            self.parentstate = None
            self.state = None
            return None
        else:
            self.parentstate, self.state = recdata
            return self.state

    def recover_parent(self):
        """Recovers the parentstate."""
        self.recover()
        return self.parentstate

    def delete(self):
        self.save(None)

    def delete_parent(self):
        """Deletes the state.  Note that this delete both the parentstate and the child state."""
        self.options.parent.delete()




"""
The functions in this module are used as heuristics to guide the search in SearchCompiler.


The required format for a heuristic is to take in a circuit, a vector of parameters for that circuit, a weight for that circuit, and an Options object, and to return a single real valued number that will be used to order the search tree.
"""
# The heuristic is a function used to guide the search.  It is called on each node after it is evaluated, and the resulting value is used to position that node's children in the search queue.

def greedy(circ, v, weight, options):
    """Defines a heuristic that results in greedy search, which focuses soley on minimizing the eval_func, and behaves somewhat similarly to depth first sarch."""
    return options.eval_func(options.target, circ.matrix(v))

def astar(circ, v, weight, options):
    """Defines a heuristic that combines the weight of the circuit with the value from eval_func.  It generally gives similar quality results to djikstra, but with a drastic reduction in the number of node evaluations."""
    return 10.0*options.eval_func(options.target, circ.matrix(v)) + weight

def djikstra(circ, v, weight, options):
    """Defines a heuristic that relies only on the weight, which gurantees a minimal-weight final solution, at the expense of a long runtime.  It behaves somewhat similarly to breadth first search."""
    return weight




"""
This module contains a list of predefined commonly used constant unitaries, and functions for generating commonly used unitaries.

Attributes:

    cnot : Constant `CNOT` unitary
    sqrt_cnot : Constant square root of `CNOT` unitary
    swap : Constant 2 qubit swap unitary
    toffoli : Constant toffoli unitary
    fredkin : Constant fredkin unitary
    peres : Constant peres unitary
    logical_or : Constant logical or unitary
    full_adder : Constant adder unitary

    rot_x : Function to generate an `X` rotation by `theta`
    rot_x_jac : Function that returns the jacobian of rot_x()
    rot_y : Function to generate an `Y` rotation by `theta`
    rot_y_jac : Function that returns the jacobian of rot_y()
    rot_z : Function to generate an `Z` rotation by `theta`
    rot_z_jac : Function that returns the jacobian of rot_z()
    qft : Returns a `n`x`n` qft matrix.
    identity : Returns a `n`x`n` identity matrix
    general_swap : Returns the swap matrix for qudits of the specified size.
    arbitrary_cnot : Returns a CNOT between any two qubits within the specified number of qubits
"""
import numpy as np

cnot = np.array([[1,0,0,0],
                  [0,1,0,0],
                  [0,0,0,1],
                  [0,0,1,0]],
                  dtype='complex128')

sqrt_cnot = np.array([[1,0,0,0],
                       [0,1,0,0],
                       [0,0,0.5+0.5j,0.5-0.5j],
                       [0,0,0.5-0.5j,0.5+0.5j]],
                       dtype='complex128')

swap = np.array([[1,0,0,0],
                  [0,0,1,0],
                  [0,1,0,0],
                  [0,0,0,1]],
                  dtype='complex128')

toffoli = np.array([[1,0,0,0,0,0,0,0],
                     [0,1,0,0,0,0,0,0],
                     [0,0,1,0,0,0,0,0],
                     [0,0,0,1,0,0,0,0],
                     [0,0,0,0,1,0,0,0],
                     [0,0,0,0,0,1,0,0],
                     [0,0,0,0,0,0,0,1],
                     [0,0,0,0,0,0,1,0]], 
                     dtype='complex128')


fredkin = np.array([[1,0,0,0,0,0,0,0],
                     [0,1,0,0,0,0,0,0],
                     [0,0,1,0,0,0,0,0],
                     [0,0,0,1,0,0,0,0],
                     [0,0,0,0,1,0,0,0],
                     [0,0,0,0,0,0,1,0],
                     [0,0,0,0,0,1,0,0],
                     [0,0,0,0,0,0,0,1]],
                     dtype='complex128')

peres = np.array([[1,0,0,0,0,0,0,0],
                   [0,1,0,0,0,0,0,0],
                   [0,0,1,0,0,0,0,0],
                   [0,0,0,1,0,0,0,0],
                   [0,0,0,0,0,0,1,0],
                   [0,0,0,0,0,0,0,1],
                   [0,0,0,0,0,1,0,0],
                   [0,0,0,0,1,0,0,0]], 
                   dtype='complex128')

logical_or = np.array([[1,0,0,0,0,0,0,0],
                        [0,1,0,0,0,0,0,0],
                        [0,0,0,1,0,0,0,0],
                        [0,0,1,0,0,0,0,0],
                        [0,0,0,0,0,1,0,0],
                        [0,0,0,0,1,0,0,0],
                        [0,0,0,0,0,0,0,1],
                        [0,0,0,0,0,0,1,0]], 
                        dtype='complex128')

full_adder = np.array([[1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
                        [0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0],
                        [0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
                        [0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0],
                        [0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0],
                        [0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0],
                        [0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0],
                        [0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0],
                        [0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0],
                        [0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0],
                        [0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0],
                        [0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0],
                        [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1],
                        [0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0],
                        [0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0],
                        [0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0]],
                        dtype='complex128')
# Full adder that converts ABC0 to ABCS.  Behavior is not well defined when the last input bit is a 1.

def rot_z(theta: float):
    return np.array([[np.exp(-1j*theta/2), 0],[0, np.exp(1j*theta/2)]], dtype='complex128')

def rot_z_jac(theta: float):
    return np.array([[-1j/2*np.exp(-1j*theta/2), 0], [0, 1j/2*np.exp(1j*theta/2)]], dtype='complex128')

def rot_x(theta: float):
    return np.array([[np.cos(theta/2), -1j*np.sin(theta/2)],[-1j*np.sin(theta/2), np.cos(theta/2)]], dtype='complex128')

def rot_x_jac(theta: float):
    return np.array([[-1/2*np.sin(theta/2), -1j/2*np.cos(theta/2)], [-1j/2*np.cos(theta/2), -1/2*np.sin(theta/2)]], dtype='complex128')

def rot_y(theta: float):
    return np.array([[np.cos(theta/2), -np.sin(theta/2)],[np.sin(theta/2), np.cos(theta/2)]], dtype='complex128')

def rot_y_jac(theta: float):
    return np.array([[-1/2*np.sin(theta/2), -1/2*np.cos(theta/2)], [1/2*np.cos(theta/2), -1/2*np.sin(theta/2)]], dtype='complex128')

def qft(n: int):
    root = np.e ** (2j * np.pi / n)
    Q = np.array(np.fromfunction(lambda x,y: root**(x*y), (n,n))) / np.sqrt(n)
    return Q

def identity(n: int): # not super necessary but saves a little code length
    return np.array(np.eye(n), dtype='complex128')

def general_swap(d: int = 2): # generates the swap matrix for qu-qudits
    f = lambda i, j: (i % d == j//d) and (i // d == j % d)
    return np.array(np.fromfunction(np.vectorize(f), (d**2, d**2)), dtype='complex128')

# generates an arbitrary cnot gate by classical logic and brute force
# it may be a good idea to write a better version of this at some point, but this should be good enough for use with the search compiler on 2-4 qubits.
def arbitrary_cnot(qudits: int, control: int, target: int):
    # this test returns if the given matrix index correspond to a "true" value of the bit at selected qubit index
    test = lambda x, value: (x // 2**(qudits-value-1)) % 2 != 0
    def f(i,j):
        # unless the control is true in both columns, just return part of the identity
        if not (test(i, control) and test(j, control)):
            return i==j
        elif test(i, target) != test(j, target):
            # if the control is true in both columns and the target is mismatched, verify everything else is matched
            for k in range(0, qudits):
                if k == target or k == control:
                    continue
                elif test(i, k) != test(j, k):
                    return 0
            return 1
        else:
            # if the control is true and matched and the target is matched, return 0
            return 0
    return np.array(np.fromfunction(np.vectorize(f), (2**qudits,2**qudits)),dtype='complex128')

